#ifndef ORB_TRANSLATOR_H
#define ORB_TRANSLATOR_H

#define NUM_LOGICAL_AXES 6
#define NUM_LOGICAL_BUTTONS 16

#define NUM_POSSIBLE_AXIS_KEYBINDINGS 6
#include <string.h>
#include <avr/pgmspace.h>

struct Axis_key_binding
{
  unsigned short axis;
  unsigned short min_value;
  unsigned short max_value;
  byte modifiers;
  byte key;
};

struct Button_key_binding
{
  unsigned short button_mask;
  unsigned char modifier;
  unsigned char key;
};

#define MOUSE_AXIS_X 1
#define MOUSE_AXIS_Y 2
#define MOUSE_AXIS_WHEEL 3

struct Axis_mouse_binding
{
  unsigned short axis;
  unsigned short mouse_axis;
  unsigned short scale;
};

struct Button_mouse_binding
{
  unsigned short button_mask;
  unsigned char mouse_button;
};

const PROGMEM unsigned char Axis_map_spaceorb_face[]= {0,1,2,3,4,5};
const PROGMEM short Polarity_spaceorb_face[] = {1,1,1,1,1,1};
const PROGMEM unsigned char Axis_map_spaceball_4k[] = {0,2,1,3,5,4};
const PROGMEM short Polarity_spaceball_4k[] = {1,-1,-1,1,-1,-1};

unsigned short sensitivity_chart( short value, const unsigned short * ptable )
{
  if ( value < 0 )
    {
      return 0;
    }
  else if ( ( value >= 0 ) && ( value <= 1023 ) )
  {
    return pgm_read_word(ptable+value);
  }
  else
  {
    return 1023;
  }
}


unsigned short 
axis_with_gain( short base, int polarity, short gain, const unsigned short * ptable )
{
  unsigned short result;
  // use polarity
  if ( polarity < 0 )
    {
      base = -base;
    }

  // use gain
  if ( gain < 0 )
    {
      base = base/(-gain);
    }
  else if ( gain > 0 )
    {
      base = base * gain;
    }

  // okay, we've scaled base by gain.  Now cap
  base = base + 512;
  if ( base < 0 )
    {
      base = 0;
    }
  if ( base > 1023 )
    {
      base = 1023;
    }
  if ( ptable != NULL )
    {
      result = sensitivity_chart( base, ptable );
    }
  else 
    {
      result = base;
    }
  return result;
}


class Orb_translator// : public Device_translator
{
public:

  bool chording;

  Button_key_binding* button_key_bindings;
  unsigned short num_button_key_bindings;
  Axis_key_binding* axis_key_bindings;
  unsigned short num_axis_key_bindings;
  Axis_mouse_binding* axis_mouse_bindings;
  unsigned short num_axis_mouse_bindings;
  Button_mouse_binding* button_mouse_bindings;
  unsigned short num_button_mouse_bindings;
  const unsigned short* p_sensitivity_table;
  short gain;
  const unsigned char *axis_map;
  // polarity is by LOGICAL axis
  const short *polarity;

  Orb_translator( void )
  {
    axis_map = Axis_map_spaceorb_face;
    polarity = Polarity_spaceorb_face;

    axis_key_bindings=NULL;
    num_axis_key_bindings = 0;
    button_key_bindings=NULL;
    num_button_key_bindings = 0;
    axis_mouse_bindings = NULL;
    num_axis_mouse_bindings = 0;
    chording = false;
    p_sensitivity_table = NULL;
    gain = 0;
  }

  void set_sensitivity_table( const unsigned short* p_new_sensitivity_table )
    {
      p_sensitivity_table = p_new_sensitivity_table;
    }

  void set_gain( short new_gain )
    {
      gain = new_gain;
    }

  void set_axis_key_bindings( Axis_key_binding* p_new_binding, int num_bindings )
    {
      axis_key_bindings = p_new_binding;
      num_axis_key_bindings = num_bindings;
    }

  void set_axis_mouse_bindings( Axis_mouse_binding* p_new_binding, int num_bindings )
    {
      axis_mouse_bindings = p_new_binding;
      num_axis_mouse_bindings = num_bindings;
    }

  void set_button_key_bindings( Button_key_binding* p_new_bindings, int num_bindings )
    {
      button_key_bindings = p_new_bindings;
      num_button_key_bindings = num_bindings;
    }

  void set_button_mouse_bindings( Button_mouse_binding* p_new_bindings, int num_bindings )
    {
      button_mouse_bindings = p_new_bindings;
      num_button_mouse_bindings = num_bindings;
    }

  void set_chording( bool new_chording )
  {
    chording = new_chording;
  }

  void set_polarity( const short* p_new_polarity )
  {
    polarity = p_new_polarity;
  }

  void set_axis_map( const unsigned char* p_new_map )
  {
    axis_map = p_new_map;
  }
  
  unsigned short scaled_axis_value( short physical_val, int polarity )
  {
    // takes a value from 0 to 1023 and returns the sensitivity chart version
    switch( polarity )
    {
    case 1 : 
      return physical_val; //sensitivity_chart( physical_val );
      break;
    case 0 :
      return 0;
      break;
    case -1 :
      return 1023-physical_val; //sensitivity_chart( 1023 - physical_val );
      break;
    }
  }

  unsigned short mapped_buttons( Logical_orb& orb ) 
  {
    unsigned short result;
    // work chording here
    if ( chording )
    {
      int chord_page = orb.physical_buttons & 3;
      result = ( orb.physical_buttons >> 2 ) << ( 4*chord_page);
    }
    else
    {
      result = orb.physical_buttons;// & 63;
    }
    return result;
  }

  unsigned short mapped_axis( Logical_orb& orb, int index )
  {
    if ( orb.orb_type == SpaceOrb360 )
      {
	return axis_with_gain( orb.axis( pgm_read_byte(&axis_map[index]) ), (int)(pgm_read_word(&polarity[index])), gain, p_sensitivity_table );
      }
    else
      {
        // for the spaceball, we have a problem-- axes can be very large and not particularly bounded (confusing).
        // this means we have to artificially bound and scale for safety reasons
        short ax = orb.axis(pgm_read_byte(&axis_map[index])) >> ( index < 3 ? 2 : 0 );
        if ( ax < -512 )
        {
          ax = -512;
        }
        else if ( ax > 511 )
        {
          ax = 511;
        }
        return axis_with_gain( ax, (int)(pgm_read_word(&polarity[index])), gain, p_sensitivity_table );
      }
  }


  void translate_keyboard_bindings( unsigned short *buffer, Logical_orb& from, Orb_device& to )
    {
      byte keypresses[7];
      //byte 0 will be modifiers; all other bytes are simultaneous keys
      memset( keypresses, 0, 7 );
      int key_cursor = 1;
      unsigned short buttons = mapped_buttons(from);
      for ( unsigned int i = 0; i < num_button_key_bindings; ++i )
	{
	  unsigned short mask =pgm_read_word( &button_key_bindings[i].button_mask );
	  if ( (buttons & mask) == mask )
	    {
	      unsigned char b = pgm_read_byte( &button_key_bindings[i].modifier );
	      if ( b != 0 )
		{
		  keypresses[0] |= b;
		}
	      b = pgm_read_byte( &button_key_bindings[i].key );
	      if ( b != 0 )
		{
		  keypresses[key_cursor] = b;
		  key_cursor++;
		}
	    }
	}

      // now add axis key mappings
       for ( unsigned short j = 0; j < num_axis_key_bindings; ++j ) 
 	{ 
	  unsigned short ax = buffer[ pgm_read_word(&axis_key_bindings[j].axis) ];
	  if ( ( key_cursor < 6 ) 
	       && ( ax >= pgm_read_word(&axis_key_bindings[j].min_value ))
	       && ( ax <= pgm_read_word(&axis_key_bindings[j].max_value )) )
	    {
	      if ( pgm_read_byte(&axis_key_bindings[j].modifiers) )
		{
		  keypresses[0] |= pgm_read_byte(&axis_key_bindings[j].modifiers);
		}
	      else
		{
		  keypresses[key_cursor] = pgm_read_byte(&axis_key_bindings[j].key);
		  key_cursor++;
		}
	    }
	}
      to.send_keyboard_report( 6, keypresses );
    }

  void translate_joystick( unsigned short *buffer, 
			   Logical_orb& from, 
			   Orb_device& to )
  {
    for ( int i = 0; i < NUM_LOGICAL_AXES; ++i )
      {
	buffer[i] = mapped_axis( from, i );
      }
    to.send_joystick_report( buffer[0],
			     buffer[1],
			     buffer[2],
			     buffer[3],
			     buffer[4],
			     buffer[5],
			     mapped_buttons(from) );
  }

  char scaled_mouse_axis( unsigned short orb_axis,
			  unsigned short scale )
    {
      short ax = ((short)(orb_axis) - 512) >> scale;
      if ( ax < -127 )
	{
	  ax = -127;
	}
      if ( ax > 127 )
	{
	  ax = 127;
	}
      return ax;
    }

  void translate_mouse_bindings( unsigned short* buffer,
				 Logical_orb& from,
				 Orb_device& to )
    {
      unsigned char mouse_buttons = 0;
      unsigned char x = 0;
      unsigned char y = 0;
      unsigned char wheel=0;
      for ( int i = 0; i < num_axis_mouse_bindings; ++i )
	{
	  unsigned short ax = buffer[ pgm_read_word(&axis_mouse_bindings[i].axis) ];
	  char mouse_ax =  scaled_mouse_axis( ax, pgm_read_word( &axis_mouse_bindings[i].scale ) );
	  switch (pgm_read_byte( &axis_mouse_bindings[i].mouse_axis ))
	    {
	    case MOUSE_AXIS_X :
	      x = mouse_ax;
	      break;
	    case MOUSE_AXIS_Y:
	      y = mouse_ax;
	      break;
	    case MOUSE_AXIS_WHEEL :
	      wheel = mouse_ax;
	      break;
	    }
	}

      // now we have a trick with mouse buttons, because if we send
      // too many button reports, it thinks we're aggressively double-clicking
      unsigned short buttons = mapped_buttons(from);
      for ( int i = 0; i < num_button_mouse_bindings; ++i )
	{
	  unsigned short mask =pgm_read_word( &button_mouse_bindings[i].button_mask );
	  if ( (buttons & mask) == mask )
	    {
	      mouse_buttons |= pgm_read_byte( &button_mouse_bindings[i].mouse_button );
	    }
	}

      // do button bindings
      to.send_mouse_report( mouse_buttons, x, y, wheel );
    }

  void update( Logical_orb& from, Orb_device& to )
  {
    unsigned short buffer[ NUM_LOGICAL_AXES ];
    translate_joystick( buffer, from, to );
    if ( (num_button_key_bindings > 0) || (num_axis_key_bindings > 0))
      {
	translate_keyboard_bindings( buffer, from, to );
      }
    if ( (num_axis_mouse_bindings > 0) || (num_button_mouse_bindings > 0) )
      {
	translate_mouse_bindings( buffer, from, to );
      }

    from.log_change();
  }
};

#endif
